/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.serialization;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.logger.Loggable;
import org.apache.cocoon.caching.CacheValidity;
import org.apache.cocoon.caching.Cacheable;
import org.apache.cocoon.caching.NOPCacheValidity;
import org.apache.cocoon.components.renderer.ExtendableRendererFactory;
import org.apache.cocoon.components.renderer.RendererFactory;
import org.apache.cocoon.util.ClassUtils;
import org.apache.fop.apps.Driver;
import org.apache.fop.apps.Options;
import org.apache.fop.render.Renderer;

import java.io.OutputStream;

/**
 * @author ?
 * @version CVS $Id: FOPSerializer.java,v 1.7 2002/03/19 09:05:56 sylvain Exp $
 */
public class FOPSerializer
extends AbstractSerializer
implements Configurable, Cacheable {

    /**
      * The <code>Options</code> used by FOP.
      */
    protected Options options;

    /**
      * The <code>Driver</code> which is FOP.
      */
    protected Driver driver;

    /**
      * The current <code>Renderer</code>.
      */
    protected Renderer renderer;

    /**
      * The Renderer Factory to use
      */
    protected static RendererFactory factory = ExtendableRendererFactory.getRendererFactoryImplementation();

    /**
      * The current <code>mime-type</code>.
      */
    protected String mimetype;

    /**
     * The renderer name if configured
     */
    protected String rendererName;
    
    /**
     * Should we set the content length ?
     */
    protected boolean setContentLength = true;

    /**
      * Set the configurations for this serializer.
      */
    public void configure(Configuration conf)
        throws ConfigurationException {
        String userConfig = null;
        java.io.File userConfigFile = null;

        if (conf != null) {
            Configuration child = conf.getChild("user-config");
            if (child != null) {
                try {
                    userConfig = child.getAttribute("src");
                } catch(Exception ex) {
                    // No config file specified
                }
            }
            
            this.setContentLength = conf.getChild("set-content-length").getValueAsBoolean(true);
        }

        // Check for null, use external or inbuilt config.
        if(userConfig != null) {
            try {
                userConfigFile = new java.io.File(userConfig);
                options = new Options(userConfigFile);
                getLogger().debug("Using config file " + userConfig);
            } catch (Exception ex) {
                getLogger().error("Cannot load  config " + userConfig, ex);
                throw new ConfigurationException("Cannot load config " + userConfig, ex);
            }
        } else {
            try {
                options = new Options();
                getLogger().debug("Using default config file");
            } catch (Exception e) {
                getLogger().error("Cannot load default config ", e);
            }
        }

        // Get the mime type.
        this.mimetype = conf.getAttribute("mime-type");
        getLogger().debug("FOPSerializer mime-type:" + mimetype);

        // Using the Renderer Factory, get the default renderer
        // for this MIME type.
        this.renderer = factory.createRenderer(mimetype);
        // Iterate through the parameters, looking for a renderer reference
        Configuration[] parameters = conf.getChildren("parameter");
        for (int i = 0; i < parameters.length; i++) {
            String name = parameters[i].getAttribute("name");
            if ("renderer".equals(name)) {
                this.rendererName = parameters[i].getAttribute("value");
                try {
                    this.renderer = (Renderer)ClassUtils.newInstance(rendererName);
                } catch (Exception ex) {
                    getLogger().error("Cannot load  class " + rendererName, ex);
                    throw new ConfigurationException("Cannot load class " + rendererName, ex);
                }
            }
        }

        // Do we have a renderer yet?
        if (this.renderer == null ) {
            throw new ConfigurationException(
                "Could not autodetect renderer for FOPSerializer and "
                + "no renderer was specified in the sitemap configuration."
            );
        }
    }

    /**
     * Return the MIME type.
     */
    public String getMimeType() {
        return mimetype;
    }

    /**
     * Create the FOP driver
     * Set the <code>OutputStream</code> where the XML should be serialized.
     */
    public void setOutputStream(OutputStream out) {
        // load the fop driver
        driver = new Driver();

        if(driver instanceof Loggable) {
            driver.setLogger(getLogger().getChildLogger("fop"));
        }
        if (this.rendererName == null) {
            this.renderer = factory.createRenderer(mimetype);
        } else {
            try {
                this.renderer = (Renderer)ClassUtils.newInstance(rendererName);
            } catch (Exception ex) {
                getLogger().error("Cannot load  class " + rendererName, ex);
                throw new RuntimeException("Cannot load class " + rendererName);
            }
        }
        driver.setRenderer(this.renderer);
        driver.setOutputStream(out);
        setContentHandler(driver.getContentHandler());
    }

    /**
     * Generate the unique key.
     * This key must be unique inside the space of this component.
     * This method must be invoked before the generateValidity() method.
     *
     * @return The generated key or <code>0</code> if the component
     *              is currently not cacheable.
     */
    public long generateKey() {
        return 1;
    }

    /**
     * Generate the validity object.
     * Before this method can be invoked the generateKey() method
     * must be invoked.
     *
     * @return The generated validity object or <code>null</code> if the
     *         component is currently not cacheable.
     */
    public CacheValidity generateValidity() {
        return NOPCacheValidity.CACHE_VALIDITY;
    }

    /**
      * Recycle serializer by removing references
      */
    public void recycle() {
        super.recycle();
        options = null;
        driver = null;
        this.renderer = null;
    }

    /**
    * Test if the component wants to set the content length
    */
    public boolean shouldSetContentLength() {
        return this.setContentLength;
    }
}
