/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.transformation;

import java.io.IOException;
import java.util.Map;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.Cacheable;
import org.apache.cocoon.caching.CacheValidity;
import org.apache.cocoon.caching.NOPCacheValidity;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.Session;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.transformation.AbstractTransformer;
import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

/**
 * The encodeURL transformer emits encoded URLs.
 * <p>
 *   This transformer applies encodeURL method to URLs.
 *   You may want to use this transform to avoid doing the manually
 *   encodeURL() call.
 * </p>
 * <p>
 *   Usually this transformer is appended as last transformer before
 *   the serialization process. In this case it is possible to encode
 *   URLs introduced in the generator, and xslt transformer phase.
 * </p>
 * <p>
 *   You can specify which attributes hold URL values in order to restrict
 *   URL rewriting to specific attributes only.
 * </p>
 * <p>
 * Usage in a sitemap:
 * </p>
 * <pre><tt>
 *   <map:composition>
 *   ...
 *     <map:transformers>
 *     ...
 *       <map:transformer type="encodeURL"
 *         src="org.apache.cocoon.optional.transformation.EncodeURLTransformer">
 *         <exclude-name>img/@src</exclude-name>
 *         <include-name>.&asterik;/@href|.&asterik;/@src|.&asterik;/@action</include-name>
 *       </map:transformer>
 *   ...
 *   <map:pipelines>
 *     <map:pipeline>
 *       ...
 *       <map:transform type="encodeURL"/>
 *       ...
 * </pre></tt>
 *
 * @author     <a href="mailto:bh22351@i-one.at">Bernhard Huber</a>
 * @version    CVS $Id: EncodeURLTransformer.java,v 1.1 2002/03/14 07:09:50 cziegeler Exp $
 *
 * @cocoon:name encodeURL
 * @cocoon:status scratchpad
 * @cocoon:parameter name="exclude-name"
 *   default-value="img/@src"
 *   description="RE pattern for excluding attributes from encode URL rewriting"
 * @cocoon:parameter name="include-name"
 *   default-value=".&asterik;/@href|.&asterik;/@src|.&asterik;/@action"
 *   description="RE pattern for including attributes from encode URL rewriting"
 *
 */

public class EncodeURLTransformer
  extends AbstractTransformer
  implements Configurable {

    /**
     * Configuration name for specifying excluding patterns,
     * ie exclude-name.
     *
     * @since
     */
    public final static String EXCLUDE_NAME = "exclude-name";
    /**
     * Configuration name for specifying including patterns,
     * ie include-name.
     *
     * @since
     */
    public final static String INCLUDE_NAME = "include-name";

    /**
     * Configuration default exclude pattern,
     * ie img/@src
     *
     * @since
     */
    public final static String EXCLUDE_NAME_DEFAULT = "img/@src";
    /**
     * Configuration default exclude pattern,
     * ie .*\/@href|.*\/@action|frame/@src
     *
     * @since
     */
    public final static String INCLUDE_NAME_DEFAULT = ".*/@href|.*/@action|frame/@src";

    private String includeNameConfigure = INCLUDE_NAME_DEFAULT;
    private String excludeNameConfigure = EXCLUDE_NAME_DEFAULT;

    private ElementAttributeMatching elementAttributeMatching;
    private Response response;
    private Session  session;


    /**
     * Setup the transformer.
     * <p>
     *   Setup include, and exclude patterns from the parameters
     * </p>
     *
     * @param resolver source resolver
     * @param objectModel sitemap objects
     * @param parameters request parameters
     *
     */
    public void setup(SourceResolver resolver, Map objectModel, String source, Parameters parameters)
    throws ProcessingException, SAXException, IOException {

        final Request request = ObjectModelHelper.getRequest(objectModel);
        this.session = request.getSession( false );
        if ( null != this.session ) {
            if ( request.isRequestedSessionIdFromURL() ) {
                this.response = ObjectModelHelper.getResponse(objectModel);
                final String includeName = parameters.getParameter(EncodeURLTransformer.INCLUDE_NAME,
                                                                   this.includeNameConfigure);
                final String excludeName = parameters.getParameter(EncodeURLTransformer.EXCLUDE_NAME,
                                                                   this.excludeNameConfigure);
                try {
                    this.elementAttributeMatching = new ElementAttributeMatching(includeName, excludeName);
                } catch (RESyntaxException reex) {
                    final String message = "Cannot parse include-name: " + includeName + " " +
                        "or exclude-name: " + excludeName + "!";
                    throw new ProcessingException(message, reex);
                }
            } else {
                // we don't need to encode
                this.session = null;
            }
        }
    }


    /**
     * BEGIN SitemapComponent methods
     *
     * @param  configuration               Description of Parameter
     * @exception  ConfigurationException  Description of Exception
     * @since
     */
    public void configure(Configuration configuration) throws ConfigurationException {
        if (configuration != null) {
            Configuration child;

            child = configuration.getChild(EncodeURLTransformer.INCLUDE_NAME);
            this.includeNameConfigure = child.getValue(INCLUDE_NAME_DEFAULT);

            child = configuration.getChild(EncodeURLTransformer.EXCLUDE_NAME);
            this.excludeNameConfigure = child.getValue(EXCLUDE_NAME_DEFAULT);
        }
        if (this.includeNameConfigure == null) {
            String message = "Configure " + INCLUDE_NAME + "!";
            throw new ConfigurationException(message);
        }
        if (this.excludeNameConfigure == null) {
            String message = "Configure " + EXCLUDE_NAME + "!";
            throw new ConfigurationException(message);
        }
    }


    /**
     *Description of the Method
     *
     * @since
     */
    public void recycle() {
        super.recycle();
        this.response = null;
        this.session = null;
        this.elementAttributeMatching = null;
    }


    /**
     * Generate the unique key.
     * This key must be unique inside the space of this component.
     *
     * @return The generated key hashes the src
     */
    public long generateKey() {
        if ( null == this.session ) {
            return 1;
        }
        return 0;
    }

    /**
     * Generate the validity object.
     *
     * @return The generated validity object or <code>null</code> if the
     *         component is currently not cacheable.
     */
    public CacheValidity generateValidity() {
        if ( null == this.session ) {
            return NOPCacheValidity.CACHE_VALIDITY;
        }
        return null;
    }

    /**
     * Start parsing an element
     *
     * @param  uri               of the element
     * @param  name              of the element
     * @param  raw               name of the element
     * @param  attributes        list
     * @exception  SAXException  Description of Exception
     * @since
     */
    public void startElement(String uri, String name, String raw, Attributes attributes)
    throws SAXException {

        if (this.session != null && this.elementAttributeMatching != null) {
            String lname = name;
            if (attributes != null && attributes.getLength() > 0) {
                AttributesImpl new_attributes = new AttributesImpl(attributes);
                for (int i = 0; i < new_attributes.getLength(); i++) {
                    String attr_lname = new_attributes.getLocalName(i);

                    String value = new_attributes.getValue(i);

                    if (elementAttributeMatching.matchesElementAttribute(lname, attr_lname)) {
                        final String new_value = response.encodeURL( value );
                        if (getLogger().isDebugEnabled()) {
                            this.getLogger().debug("element/@attribute matches: " + name + "/@" + attr_lname);
                            this.getLogger().debug("encodeURL: " + value + " -> " + new_value);
                        }
                        new_attributes.setValue(i, new_value);
                    }
                }
                // parent handles element using encoded attribute values
                super.contentHandler.startElement(uri, name, raw, new_attributes);
                return;
            }
        }
        // no match, parent handles element as-is
        super.contentHandler.startElement(uri, name, raw, attributes);
    }

    /**
     * A helper class for matching element names, and attribute names.
     *
     * <p>
     *  For given include-name, exclude-name decide if element-attribute pair
     *  matches. This class defines the precedence and matching algorithm.
     * </p>
     *
     * @author     <a href="mailto:bh22351@i-one.at">Bernhard Huber</a>
     * @version    CVS $Id: EncodeURLTransformer.java,v 1.1 2002/03/14 07:09:50 cziegeler Exp $
     */
    public class ElementAttributeMatching {
        /**
         * Regular expression of including patterns
         *
         * @since
         */
        protected RE includeNameRE;
        /**
         * Regular expression of excluding patterns
         *
         * @since
         */
        protected RE excludeNameRE;


        /**
         *Constructor for the ElementAttributeMatching object
         *
         * @param  includeName            Description of Parameter
         * @param  excludeName            Description of Parameter
         * @exception  RESyntaxException  Description of Exception
         * @since
         */
        public ElementAttributeMatching(String includeName, String excludeName) throws RESyntaxException {
            includeNameRE = new RE(includeName, RE.MATCH_CASEINDEPENDENT);
            excludeNameRE = new RE(excludeName, RE.MATCH_CASEINDEPENDENT);
        }


        /**
         * Return true iff element_name attr_name pair is not matched by exclude-name,
         * but is matched by include-name
         *
         * @param  element_name
         * @param  attr_name
         * @return               boolean true iff value of attribute_name should get rewritten, else
         *   false.
         * @since
         */
        public boolean matchesElementAttribute(String element_name, String attr_name) {
            String element_attr_name = canonicalizeElementAttribute(element_name, attr_name);

            if (excludeNameRE != null && includeNameRE != null) {
                return !matchesExcludesElementAttribute(element_attr_name) &&
                        matchesIncludesElementAttribute(element_attr_name);
            } else {
                return false;
            }
        }


        /**
         * Build from elementname, and attribute name a single string.
         * <p>
         *   String concatenated <code>element name + "/@" + attribute name</code>
         *   is matched against the include and excluding patterns.
         * </p>
         *
         * @param  element_name  Description of Parameter
         * @param  attr_name     Description of Parameter
         * @return               Description of the Returned Value
         * @since
         */
        String canonicalizeElementAttribute(String element_name, String attr_name) {
            StringBuffer element_attr_name = new StringBuffer();
            element_attr_name.append(element_name);
            element_attr_name.append("/@");
            element_attr_name.append(attr_name);
            return element_attr_name.toString();
        }


        /**
         * Return true iff element_name attr_name pair is matched by exclude-name.
         *
         * @param  element_attr_name
         * @return                    boolean true iff exclude-name matches element_name, attr_name, else
         *   false.
         * @since
         */
        private boolean matchesExcludesElementAttribute(String element_attr_name) {
            boolean match = excludeNameRE.match(element_attr_name);
            return match;
        }


        /**
         * Return true iff element_name attr_name pair is matched by include-name.
         *
         * @param  element_attr_name
         * @return                    boolean true iff include-name matches element_name, attr_name, else
         *   false.
         * @since
         */
        private boolean matchesIncludesElementAttribute(String element_attr_name) {
            boolean match = includeNameRE.match(element_attr_name);
            return match;
        }
    }
}

