package org.apache.cocoon.components.flow.javascript;

import java.util.HashMap;
import java.util.Map;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.cocoon.components.flow.AbstractInterpreter;
import org.apache.cocoon.components.flow.ContinuationsManager;
import org.apache.cocoon.components.flow.ContinuationsManagerImpl;
import org.apache.cocoon.components.flow.WebContinuation;
import org.apache.cocoon.components.treeprocessor.InvokeContext;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.mozilla.javascript.ScriptRuntime;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import org.mozilla.javascript.Undefined;
import org.mozilla.javascript.Wrapper;

/**
 * JavaScript interface to various Cocoon abstractions.
 *
 * @author <a href="mailto:ovidiu@cup.hp.com">Ovidiu Predescu</a>
 * @since March 16, 2002
 */
public class JSCocoon extends ScriptableObject
{
  protected AbstractInterpreter interp;
  protected Scriptable scope;
  protected Environment environment;
  protected InvokeContext context;
  protected ComponentManager manager;
  protected ContinuationsManager continuationsMgr;

  public JSCocoon() {}

  public String getClassName()
  {
    return "Cocoon";
  }

  public void setScope(Scriptable scope)
  {
    this.scope = scope;
  }

  public Scriptable getScope()
  {
    return scope;
  }

  public void setInterpreter(AbstractInterpreter interp)
  {
    this.interp = interp;
  }

  public void setContext(ComponentManager manager, Environment environment,
                         InvokeContext ctx)
  {
    this.manager = manager;
    this.environment = environment;
    this.context = ctx;
  }

  public void invalidateContext()
  {
    manager = null;
    environment = null;
    context = null;
  }

  public void setContinuationsManager(ContinuationsManager mgr)
  {
    this.continuationsMgr = mgr;
  }

  ContinuationsManager getContinuationsManager()
  {
    return continuationsMgr;
  }

  public AbstractInterpreter jsGet_interpreter()
  {
    return interp;
  }

  public Environment jsGet_environment()
  {
    return environment;
  }

  public InvokeContext jsGet_invokeContext()
  {
    return context;
  }

  public Request jsGet_request()
  {
    Map objectModel = environment.getObjectModel();
    return ObjectModelHelper.getRequest(objectModel);
  }

  public Response jsGet_response()
  {
    Map objectModel = environment.getObjectModel();
    return ObjectModelHelper.getResponse(objectModel);
  }

  public Context jsGet_context()
  {
    Map objectModel = environment.getObjectModel();
    return ObjectModelHelper.getContext(objectModel);
  }

  /**
   * Load the file specified as argument. Registers the file with the
   * interpreter and then forces its loading by calling {@link
   * AbstractInterpreter#checkForModifiedScripts}.
   *
   * @param filename a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception Exception if an error occurs
   */
  public Object jsFunction_load(String filename)
    throws Exception
  {
    try {
      interp.register(filename);
      interp.checkForModifiedScripts(environment);
    }
    catch (Exception ex) {
      ex.printStackTrace();
      throw ex;
    }
    return null;
  }

  public String jsFunction_toString()
  {
    return "[object " + toString() + "]";
  }

  public void jsFunction_processPipeline(String name, Object pipelineArgs,
                                         Object bizData,
                                         Object cont)
    throws Exception
  {
    if (pipelineArgs instanceof Wrapper)
      pipelineArgs = ((Wrapper)pipelineArgs).unwrap();
    else if (pipelineArgs instanceof Scriptable)
      pipelineArgs = jsobjectToMap((Scriptable)pipelineArgs);

    if (bizData instanceof Wrapper)
      bizData = ((Wrapper)bizData).unwrap();
    else if (bizData instanceof Scriptable)
      bizData = jsobjectToMap((Scriptable)bizData);

    JSWebContinuation kont = (JSWebContinuation)cont;

    interp.processPipeline(name, (Map)pipelineArgs, bizData,
                           kont.getWebContinuation(),
                           environment, context);
  }

  public void jsFunction_diplayAllContinuations()
  {
    if (continuationsMgr instanceof ContinuationsManagerImpl)
      ((ContinuationsManagerImpl)continuationsMgr).displayAllContinuations();
  }

  // All right, this breaks the encapsulation, but I couldn't
  // find any better way to obtain the ComponentManager from a
  // JSWebContinuation.
  ComponentManager getComponentManager()
  {
    return manager;
  }

  public static Map jsobjectToMap(Scriptable jsobject)
  {
    HashMap hash = new HashMap();
    Object[] ids = jsobject.getIds();
    for (int i = 0; i < ids.length; i++) {
      String key = ScriptRuntime.toString(ids[i]);
      Object value = jsobject.get(key, jsobject);
      if (value == Undefined.instance)
        value = null;
      else
        value = ScriptRuntime.toPrimitive(value);
      hash.put(key, value);
    }
    return hash;
  }
}
