/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.sunshine.connector;

import java.io.*;
import java.net.*;
import java.util.*;

import org.apache.avalon.excalibur.pool.Recyclable;
import org.apache.avalon.excalibur.source.SourceParameters;
import org.apache.avalon.excalibur.source.SourceUtil;
import org.apache.avalon.framework.component.Composable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.component.ComponentSelector;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.logger.AbstractLoggable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.activity.Disposable;

import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.Processor;
import org.apache.cocoon.ResourceNotFoundException;
import org.apache.cocoon.components.RequestLifecycleComponent;
import org.apache.cocoon.components.parser.Parser;
import org.apache.cocoon.components.source.URLSource;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.Source;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.xml.XMLConsumer;
import org.apache.cocoon.xml.dom.DOMBuilder;
import org.apache.cocoon.sunshine.helpers.URLRewriter;
import org.apache.cocoon.xml.IncludeXMLConsumer;
import org.apache.cocoon.sunshine.xml.XMLUtil;

import org.w3c.dom.*;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.*;

/**
 * The Component for loading and saving xml to external resource connectors.
 *
 * @author <a href="mailto:cziegeler@s-und-n.de">Carsten Ziegeler</a>
 * @version CVS $Id: ResourceConnectorImpl.java,v 1.3 2002/03/11 11:15:25 cziegeler Exp $
*/
public final class ResourceConnectorImpl
extends AbstractLoggable
implements Recyclable, ResourceConnector,
           Composable, RequestLifecycleComponent {

    /** The <code>ComponentManager</code> */
    protected ComponentManager manager;

    /** Empty attributes */
    protected Attributes emptyAttributes = new AttributesImpl();

    protected Map objectModel;

    protected SourceResolver resolver;

    /**
     * Composer
     */
    public void compose(ComponentManager manager)
    throws ComponentException {
        this.manager = manager;
    }

    /**
     * Set the <code>SourceResolver</code>, objectModel <code>Map</code>,
     * used to process the request.
     *  Set up the sunShine component.
     *  This method is automatically called for each request. Do not invoke
     *  this method by hand.
     */
    public void setup(SourceResolver resolver, Map objectModel)
    throws ProcessingException, SAXException, IOException {
        this.objectModel = objectModel;
        this.resolver = resolver;
    }

    /**
     * Recycle
     */
    public void recycle() {
        this.objectModel = null;
        this.resolver = null;
    }

    private static String[] outputKeys = new String[] {
        javax.xml.transform.OutputKeys.CDATA_SECTION_ELEMENTS,
        javax.xml.transform.OutputKeys.DOCTYPE_PUBLIC,
        javax.xml.transform.OutputKeys.DOCTYPE_SYSTEM,
        javax.xml.transform.OutputKeys.ENCODING,
        javax.xml.transform.OutputKeys.INDENT,
        javax.xml.transform.OutputKeys.MEDIA_TYPE,
        javax.xml.transform.OutputKeys.METHOD,
        javax.xml.transform.OutputKeys.OMIT_XML_DECLARATION,
        javax.xml.transform.OutputKeys.STANDALONE,
        javax.xml.transform.OutputKeys.VERSION};

    /**
     * Set OutputKeys from parameters
     */
    private void setOutputKeys(Properties properties, Parameters parameters) {
        if (properties != null && parameters != null) {
            for(int i = 0; i < outputKeys.length; i++) {
                final String value = parameters.getParameter(outputKeys[i], null);
                if (value != null) {
                    properties.setProperty(outputKeys[i], value);
                }
            }
        }
    }

    /**
     * Get a <code>Source</code> object
     */
    private Source getSource(String             uri,
                             Parameters         typeParameters,
                             SourceParameters resourceParameters)
    throws IOException, SAXException, ProcessingException {
        Source input;

        // Test: local uri (= same servlet/cocoon) ?
        if (uri.startsWith("/") == true) {
            // server-absolute url is transformed to absolute url
            Request request = ObjectModelHelper.getRequest(objectModel);
            uri = request.getScheme() + "://" +
                  request.getServerName() + ":" + request.getServerPort() + uri;
        }

        // first step: encode parameters which are already appended to the url
        int queryPos = uri.indexOf('?');
        if (queryPos != -1) {
            String queryString = uri.substring(queryPos+1);
            SourceParameters queries = new SourceParameters(queryString);
            if (queries.hasParameters()) {
                StringBuffer buffer;
                buffer = new StringBuffer(uri.substring(0, queryPos));
                String current;
                Iterator iter = queries.getParameterNames();
                char separator = '?';
                Iterator values;
                while (iter.hasNext() == true) {
                    current = (String)iter.next();
                    values = queries.getParameterValues(current);
                    while (values.hasNext()) {
                        buffer.append(separator)
                              .append(current)
                              .append('=')
                              .append(SourceUtil.encode((String)values.next()));
                        separator = '&';
                    }
                }
                uri = buffer.toString();
            }
        }
        boolean followRedirects = (typeParameters != null ?
                                   typeParameters.getParameterAsBoolean("followRedirects", true)
                                      : true);
        String method = (typeParameters != null ? typeParameters.getParameter("method", "GET")
                                                     : "GET");
        if (method.equalsIgnoreCase("POST") == true
            && (resourceParameters == null || resourceParameters.hasParameters() == false)) {
            method = "GET";
        }
        if ((method.equalsIgnoreCase("POST") == false
             || uri.startsWith("cocoon:") == true)
            && resourceParameters != null
            && resourceParameters.hasParameters()) {

            int pos = uri.indexOf(";jsessionid=");
            if (uri.startsWith("cocoon:") == false) {
                if (pos != -1) uri = uri.substring(0, pos);
                uri = SourceUtil.appendParameters(uri, resourceParameters);
            } else {
                StringBuffer buf;
                if (pos == -1) {
                    buf = new StringBuffer(uri);
                } else {
                    buf = new StringBuffer(uri.substring(0, pos));
                }
                buf.append((uri.indexOf('?') == -1 ? '?' : '&'));
                buf.append(resourceParameters.getEncodedQueryString());
                uri = buf.toString();
            }
        }
        if (this.getLogger().isInfoEnabled() == true) {
            this.getLogger().info("Getting content of '"+uri+"' using method " + method);
        }
        input = this.resolver.resolve(uri);
        if (method.equalsIgnoreCase("POST") == true
            && input instanceof URLSource) {
            ((URLSource)input).setPostParameters(resourceParameters);
        }
        if (followRedirects == false && input instanceof URLSource) {
            ((URLSource)input).setFollowRedirects(false);
        }

        return input;
    }

    /**
     * Load XML
     */
    private DocumentFragment loadXMLFromURI(Parameters        typeParameters,
                                           String             uri,
                                           SourceParameters resourceParameters)
    throws IOException, SAXException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN loadXMLFromURI uri=" + uri +
                              ", typeParams="+typeParameters+
                              ", parameters="+resourceParameters);
        }
        if (uri == null || uri.trim().length() == 0) {
            throw new IllegalArgumentException("loadXMLFromURI: URI is required");
        }

        long startTime = System.currentTimeMillis();

        DocumentFragment recordedDocFrag = null;
        DOMBuilder       builder;

        builder = new DOMBuilder();
        builder.startDocument();
        builder.startElement("", "sunShine", "sunShine", this.emptyAttributes);

        IncludeXMLConsumer filter;
        // Test for url rewriting
        if (typeParameters != null
            && typeParameters.getParameter(URLRewriter.PARAMETER_MODE, null) != null) {
            XMLConsumer consumer = new URLRewriter(typeParameters,
                                                   builder,
                                                   builder);
            filter = new IncludeXMLConsumer(consumer, consumer);
        } else {
            filter = new IncludeXMLConsumer(builder, builder);
        }
        Source input = null;
        try {
            input = this.getSource(uri, typeParameters, resourceParameters);

            if (input != null) {
                input.toSAX(filter);
            }
        } finally {
            if (input != null) input.recycle();
            input = null;
        }

        builder.endElement("", "sunShine", "sunShine");
        builder.endDocument();

        // Create Document Fragment
        final Document doc = builder.getDocument();
        recordedDocFrag = doc.createDocumentFragment();
        final Node root = doc.getDocumentElement();
        Node child;
        while (root.hasChildNodes() == true) {
            child = root.getFirstChild();
            child.normalize();
            root.removeChild(child);
            recordedDocFrag.appendChild(child);
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END loadXMLFromURI fragment="+(recordedDocFrag == null ? "null" : XMLUtil.serializeNodeToXML(recordedDocFrag)));
        }
        return recordedDocFrag;
    }

    /**
     * Save XML to an URI. The fragment is passed to the resource as the value
     * of the parameter "content".
     */
    private void saveXMLToURI(Parameters        typeParameters,
                              String             uri,
                              SourceParameters resourceParameters,
                              DocumentFragment   fragment)
    throws IOException, SAXException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN saveXMLToURI fragment="+(fragment == null ? "null" : XMLUtil.serializeNodeToXML(fragment))+
                              ", typeParams="+typeParameters+
                              ", uri=" + uri +
                              ", parameters="+resourceParameters);
        }
        if (uri == null || uri.trim().length() == 0) {
            throw new IllegalArgumentException("saveXMLToURI: URI is required");
        }

        Properties format = XMLUtil.defaultSerializeToXMLFormat(true);
        this.setOutputKeys(format, typeParameters);
        String content = XMLUtil.serializeNode(fragment, format);
        if (resourceParameters == null) {
            resourceParameters = new SourceParameters();
        } else {
            resourceParameters = (SourceParameters)resourceParameters.clone();
        }
        resourceParameters.setSingleParameterValue("content", content);

        // The trick here is that we do a simple loadXML and ignore the result
        this.loadXMLFromURI(typeParameters, uri, resourceParameters);

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END saveXMLToURI");
        }
    }

    /**
     * Save XML to an URI. The fragment is passed to the resource as the value
     * of the parameter "content".
     */
    private void saveContentToURI(Parameters         typeParameters,
                                  String             uri,
                                  SourceParameters resourceParameters,
                                  String             content)
    throws IOException, SAXException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN saveContentToURI content="+content+
                              ", typeParams="+typeParameters+
                              ", uri=" + uri +
                              ", parameters="+resourceParameters);
        }
        if (uri == null || uri.trim().length() == 0) {
            throw new IllegalArgumentException("saveContentToURI: URI is required");
        }

        if (resourceParameters == null) {
            resourceParameters = new SourceParameters();
        } else {
            resourceParameters = (SourceParameters)resourceParameters.clone();
        }
        resourceParameters.setSingleParameterValue("content", content);

        // The trick here is that we do a simple loadXML and ignore the result
        this.loadXMLFromURI(typeParameters, uri, resourceParameters);

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END saveContentToURI");
        }
    }

    /**
     * Load XML by using an XMLLoader
     */
    private DocumentFragment loadXMLFromClass(Parameters         typeParameters,
                                             String             className,
                                             SourceParameters resourceParameters)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN loadXMLFromClass class=" + className+
                              ", typeParams="+typeParameters+
                              ", parameters="+resourceParameters);
        }
        DocumentFragment frag;

        try {
            Class loaderClass = Class.forName(className);
            XMLLoader loader = (XMLLoader)loaderClass.newInstance();

            frag = loader.load(this.objectModel, resourceParameters);
        } catch (ClassNotFoundException cnfException) {
            throw new ProcessingException("loadXMLFromClass: Class not found: " + className, cnfException);
        } catch (IllegalAccessException iaException) {
            throw new ProcessingException("loadXMLFromClass: Illegal Access: " + className, iaException);
        } catch (InstantiationException iException) {
            throw new ProcessingException("loadXMLFromClass: Instantion exception: " + className, iException);
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END loadXMLFromClass fragment="+(frag == null ? "null" : XMLUtil.serializeNodeToXML(frag)));
        }
        return frag;
    }

    /**
     * Save XML by using an XMLSaver
     */
    private void saveXMLToClass(Parameters         typeParameters,
                               String             className,
                               SourceParameters resourceParameters,
                               DocumentFragment   fragment)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN saveXMLToClass fragment="+(fragment == null ? "null" : XMLUtil.serializeNodeToXML(fragment))+
                              ", typeParams="+typeParameters+
                              ", class=" + className+
                              ", params="+resourceParameters);
        }
        try {
            Class saverClass = Class.forName(className);
            XMLSaver saver = (XMLSaver)saverClass.newInstance();

            saver.save(fragment, this.objectModel, resourceParameters);
        } catch (ClassNotFoundException cnfException) {
            throw new ProcessingException("saveXMLToClass: Class not found: " + className, cnfException);
        } catch (IllegalAccessException iaException) {
            throw new ProcessingException("saveXMLToClass: Illegal Access: " + className, iaException);
        } catch (InstantiationException iException) {
            throw new ProcessingException("saveXMLToClass: Instantion exception: " + className, iException);
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END saveXMLToClass");
        }
    }

    /**
     * Load XML from a file
     */
    private DocumentFragment loadXMLFromFile(Parameters         typeParameters,
                                            String             fileName,
                                            SourceParameters resourceParameters)
    throws ProcessingException, IOException, SAXException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN loadXMLFromFile fileName=" + fileName+
                              ", typeParams="+typeParameters+
                              ", parameters="+resourceParameters);
        }
        DocumentFragment frag = null;

        long startTime = System.currentTimeMillis();

        Source input = null;
        try {
            input = this.resolver.resolve(fileName);

            DOMBuilder builder = new DOMBuilder();
            builder.startDocument();
            builder.startElement("", "sunShine", "sunShine", this.emptyAttributes);

            IncludeXMLConsumer filter = new IncludeXMLConsumer(builder, builder);
            input.toSAX(filter);

            builder.endElement("", "sunShine", "sunShine");
            builder.endDocument();

            // Create Document Fragment
            final Document doc = builder.getDocument();
            frag = doc.createDocumentFragment();
            final Node root = doc.getDocumentElement();
            Node child;
            while (root.hasChildNodes() == true) {
                child = root.getFirstChild();
                child.normalize();
                root.removeChild(child);
                frag.appendChild(child);
            }
        } finally {
            if (input != null) input.recycle();
            input = null;
        }


        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END loadXMLFromFile fragment="+(frag == null ? "null" : XMLUtil.serializeNodeToXML(frag)));
        }
        return frag;
    }

    /**
     * Save XML to a file
     */
    private void saveXMLToFile(Parameters         typeParameters,
                              String             fileName,
                              SourceParameters resourceParameters,
                              DocumentFragment   fragment)
    throws SAXException, IOException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN saveXMLToFile fragment="+(fragment == null ? "null" : XMLUtil.serializeNodeToXML(fragment))+
                              ", typeParams="+typeParameters+
                              ", file=" + fileName +
                              ", params="+resourceParameters);
        }
        try {
            // cleanup fragment
            NodeList childs = fragment.getChildNodes();
            if (childs != null) {
                Node[] list = new Node[childs.getLength()];
                for(int i = 0; i < childs.getLength(); i++) {
                    list[i] = childs.item(i);
                }

                for(int i = 0; i < list.length; i++) {
                    if (list[i].getNodeType() == Node.TEXT_NODE) {
                        String value = list[i].getNodeValue();
                        if (value == null
                            || value.length() == 0
                            || value.trim().length() == 0) {
                            list[i].getParentNode().removeChild(list[i]);
                        }
                    }
                }
            }

            Source input = null;
            try {
                Properties format = XMLUtil.defaultSerializeToXMLFormat(false);
                this.setOutputKeys(format, typeParameters);
                input = this.resolver.resolve(fileName);
                String absolutePath = input.getSystemId();
                if (absolutePath.startsWith("file:") == false) {
                    throw new ProcessingException("Saving to " + fileName + " is not possible.");
                }
                File file = new File(absolutePath.substring("file:".length()));
                if (file.exists() == false) {
                    File directory = file.getParentFile();
                    if (directory.exists() == false) {
                        directory.mkdirs();
                    }
                    file.createNewFile();
                }
                String xml = XMLUtil.serializeNode(fragment, format);
                Writer writer = new BufferedWriter(new FileWriter(file));;
                writer.write(xml);
                writer.flush();
                writer.close();
                writer = null;
            } finally {
                if (input != null) input.recycle();
                input = null;
            }

        } catch (IOException ioe) {
            throw new ProcessingException("saveXMLToFile: IOException: " + ioe, ioe);
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END saveXMLToFile");
        }
    }

    /**
     * Save XML to a file
     */
    private void saveContentToFile(Parameters         typeParameters,
                                   String             fileName,
                                   SourceParameters resourceParameters,
                                   String             content)
    throws SAXException, IOException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN saveContentToFile content="+content+
                              ", typeParams="+typeParameters+
                              ", file=" + fileName +
                              ", params="+resourceParameters);
        }
        try {
            Source input = null;
            try {
                input = this.resolver.resolve(fileName);
                String absolutePath = input.getSystemId();
                if (absolutePath.startsWith("file:") == false) {
                    throw new ProcessingException("Saving to " + fileName + " is not possible.");
                }
                File file = new File(absolutePath.substring("file:".length()));
                if (file.exists() == false) {
                    File directory = file.getParentFile();
                    if (directory.exists() == false) {
                        directory.mkdirs();
                    }
                    file.createNewFile();
                }
                Writer writer = new BufferedWriter(new FileWriter(file));
                writer.write(content);
                writer.flush();
                writer.close();
                writer = null;
            } finally {
                if (input != null) input.recycle();
                input = null;
            }

        } catch (IOException ioe) {
            throw new ProcessingException("saveContentToFile: IOException: " + ioe, ioe);
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END saveContentToFile");
        }
    }

    /**
     * Load an XML Fragment from a resource
     */
    public DocumentFragment loadXML(int                resourceType,
                                    Parameters         typeParameters,
                                    String             resourceIdentifier,
                                    SourceParameters resourceParameters)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("start loading '"+resourceIdentifier+"'");
        }
        DocumentFragment frag = null;
        try {
            switch (resourceType) {
                case ResourceConnector.RESOURCE_TYPE_URI: frag = this.loadXMLFromURI(typeParameters,
                    resourceIdentifier, resourceParameters);
                    break;
               case ResourceConnector.RESOURCE_TYPE_CLASS: frag = this.loadXMLFromClass(typeParameters,
                    resourceIdentifier, resourceParameters);
                    break;
               case ResourceConnector.RESOURCE_TYPE_FILE: frag = this.loadXMLFromFile(typeParameters,
                    resourceIdentifier, resourceParameters);
                    break;
               default: throw new ProcessingException("ResourceConnector: type " +
                          resourceType + " not supported.");
            }
        } catch (SAXException ioe) {
            throw new ProcessingException("SAXException: " + ioe, ioe);
        } catch (IOException ioe) {
            throw new ProcessingException("IOException: " + ioe, ioe);
        }
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("finished loading '"+resourceIdentifier+"'");
        }
        return frag;
    }

    /**
     * Save an XML Fragment to a resource
     */
    public void saveXML(int                resourceType,
                        Parameters         typeParameters,
                        String             resourceIdentifier,
                        SourceParameters resourceParameters,
                        DocumentFragment   fragment)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("start saving '"+resourceIdentifier+"'");
        }
        try {
            switch (resourceType) {
                case ResourceConnector.RESOURCE_TYPE_URI: this.saveXMLToURI(typeParameters,
                           resourceIdentifier, resourceParameters, fragment);
                       break;
                case ResourceConnector.RESOURCE_TYPE_CLASS: this.saveXMLToClass(typeParameters,
                           resourceIdentifier, resourceParameters, fragment);
                       break;
                case ResourceConnector.RESOURCE_TYPE_FILE: this.saveXMLToFile(typeParameters,
                           resourceIdentifier, resourceParameters, fragment);
                       break;
                default: throw new ProcessingException("ResourceConnector: type " +
                           resourceType + " not supported.");
            }
        } catch (SAXException ioe) {
            throw new ProcessingException("SAXException: " + ioe, ioe);
        } catch (IOException ioe) {
            throw new ProcessingException("IOException: " + ioe, ioe);
        }
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("finished saving '"+resourceIdentifier+"'");
        }
    }

    /**
     * Stream XML from a file
     */
    private void streamXMLFromFile(Parameters         typeParameters,
                                   String             fileName,
                                   SourceParameters resourceParameters,
                                   org.xml.sax.ContentHandler contentHandler,
                                   LexicalHandler             lexicalHandler)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("BEGIN streamXMLFromFile fileName=" + fileName +
                              ", typeParams="+typeParameters+
                              ", parameters="+resourceParameters);
        }
        long startTime = System.currentTimeMillis();
        Source input = null;
        try {
            input = this.resolver.resolve(fileName);

            IncludeXMLConsumer filter = new IncludeXMLConsumer(contentHandler, lexicalHandler);
            input.toSAX(filter);

        } catch (SAXException sax) {
            throw new ProcessingException("SAXException: " + sax, sax);
        } catch (IOException ioe) {
            throw new ProcessingException("IOException: " + ioe, ioe);
        } finally {
            if (input != null) input.recycle();
            input = null;
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END streamXMLFromFile");
        }
    }

    /**
     * Stream XML
     */
    private void streamXMLFromURI(Parameters         typeParameters,
                                  String             uri,
                                  SourceParameters resourceParameters,
                                  org.xml.sax.ContentHandler contentHandler,
                                  LexicalHandler             lexicalHandler)
    throws IOException, SAXException, ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
           this.getLogger().debug("BEGIN streamXMLFromURI uri=" + uri +
                             ", typeParams="+typeParameters+
                             ", parameters="+resourceParameters+
                             ", contentHandler="+contentHandler+
                             ", lexicalHandler="+lexicalHandler);
        }

        if (uri == null || uri.trim().length() == 0) {
            throw new IllegalArgumentException("streamXMLFromURI: URI is required");
        }

        long startTime = System.currentTimeMillis();
        // Test for url rewriting
        if (typeParameters != null
            && typeParameters.getParameter(URLRewriter.PARAMETER_MODE, null) != null) {
            XMLConsumer consumer = new URLRewriter(typeParameters,
                                                   contentHandler,
                                                   lexicalHandler);
            contentHandler = consumer;
            lexicalHandler = consumer;
        }
        IncludeXMLConsumer filter = new IncludeXMLConsumer(contentHandler, lexicalHandler);

        Source input = null;
        try {
            input = this.getSource(uri, typeParameters, resourceParameters);
            input.toSAX(filter);
        } finally {
            if (input != null) input.recycle();
            input = null;
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END streamXMLFromURI");
        }
    }

    /**
     * Stream an XML resource.
     * This is like loading the resource and then streaming it, but without
     * the Resource->SAX->DOM->SAX transformation, its simply: Resource->SAX
     */
    public void streamXML(int                resourceType,
                          Parameters         typeParameters,
                          String             resourceIdentifier,
                          SourceParameters resourceParameters,
                          org.xml.sax.ContentHandler contentHandler,
                          LexicalHandler             lexicalHandler)
    throws ProcessingException {
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("start streaming '"+resourceIdentifier+"'");
        }
        try {
            switch (resourceType) {
            case ResourceConnector.RESOURCE_TYPE_URI: this.streamXMLFromURI(typeParameters,
               resourceIdentifier, resourceParameters, contentHandler, lexicalHandler);
               break;
            case ResourceConnector.RESOURCE_TYPE_CLASS:
                 throw new ProcessingException("streamXML: Type class not implemented");
                 // break;
            case ResourceConnector.RESOURCE_TYPE_FILE: this.streamXMLFromFile(typeParameters,
               resourceIdentifier, resourceParameters, contentHandler, lexicalHandler);
               break;

            default: throw new ProcessingException("ResourceConnector: type " +
               resourceType + " not supported.");
            }
        } catch (SAXException ioe) {
            throw new ProcessingException("SAXException: " + ioe, ioe);
        } catch (IOException ioe) {
            throw new ProcessingException("IOException: " + ioe, ioe);
        }
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("finished streaming '"+resourceIdentifier+"'");
        }
    }


    /**
     * Get the content of a uri
     */
    private String getContentFromURI(Parameters         typeParameters,
                                    String             uri,
                                    SourceParameters resourceParameters)
    throws SAXException, ProcessingException, IOException {
        if (this.getLogger().isDebugEnabled() == true) {
           this.getLogger().debug("BEGIN getContentFromURI uri=" + uri +
                             ", typeParams="+typeParameters+
                             ", parameters="+resourceParameters);
        }
        String content = null;

        if (uri == null || uri.trim().length() == 0) {
            throw new IllegalArgumentException("getContentFromURI: URI is required");
        }

        long startTime = System.currentTimeMillis();
        Source input = null;
        try {
            input = this.getSource(uri, typeParameters, resourceParameters);
            InputSource source = input.getInputSource();
            InputStream stream = source.getByteStream();
            StringBuffer buffer = new StringBuffer();
            String encoding = source.getEncoding();
            int available;
            byte[] data;
            do {
                available = 1024;
                data = new byte[available];
                available = stream.read(data, 0, available);
                if (available > 0) {
                    if (encoding == null) {
                        buffer.append(new String(data, 0, available));
                    } else {
                        buffer.append(new String(data, 0, available, encoding));
                    }
                }
            } while (available > 0);
            content = buffer.toString();
        } finally {
            if (input != null) input.recycle();
            input = null;
        }

        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END getContentFromURI content="+content);
        }
        return content;
    }

    /**
     * Get the content of a file
     */
    private String getContentFromFile(Parameters         typeParameters,
                                      String             filename,
                                      SourceParameters resourceParameters)
    throws SAXException, ProcessingException, IOException {
        if (this.getLogger().isDebugEnabled() == true) {
           this.getLogger().debug("BEGIN getContentFromFile filename=" + filename +
                             ", typeParams="+typeParameters+
                             ", parameters="+resourceParameters);
        }
        String content;
        Source input = null;
        try {
            input = this.resolver.resolve(filename);
            String absolutePath = input.getSystemId();
            if (absolutePath.startsWith("file:") == false) {
                throw new ProcessingException("Saving to " + filename + " is not possible.");
            }
            File file = new File(absolutePath.substring("file:".length()));
            byte[] data = null;
            InputStream fis = new FileInputStream(file);
            int available;
            byte[] tempData;
            byte[] copyData;
            do {
                available = 1024;
                tempData = new byte[available];
                available = fis.read(tempData, 0, available);
                if (available > 0) {
                    copyData = new byte[(data == null ? 0 : data.length) + available];
                    if (data != null) {
                        System.arraycopy(data, 0, copyData, 0, data.length);
                    }
                    System.arraycopy(tempData, 0, copyData, (data == null ? 0 : data.length), available);
                    data = copyData;
                }
            } while (available > 0);
            fis.close();
            content = (data == null ? "" : new String(data));
        } catch (FileNotFoundException local) {
            throw new ResourceNotFoundException("File not found '" + filename + "'", local);
        } finally {
            if (input != null) input.recycle();
            input = null;
        }
        if (this.getLogger().isDebugEnabled() == true) {
            this.getLogger().debug("END getContentFromFile content="+content);
        }
        return content;
    }

    /**
     * Get the content
     */
    public String getContent(int                resourceType,
                             Parameters         typeParameters,
                             String             resourceIdentifier,
                             SourceParameters resourceParameters)
    throws ProcessingException {
        String content = null;

        try {
            switch (resourceType) {
                case ResourceConnector.RESOURCE_TYPE_URI: content = this.getContentFromURI(typeParameters,
                    resourceIdentifier, resourceParameters);
                    break;
               case ResourceConnector.RESOURCE_TYPE_CLASS:
                    throw new ProcessingException("getContent: Type class not implemented.");
               case ResourceConnector.RESOURCE_TYPE_FILE: content = this.getContentFromFile(typeParameters,
                    resourceIdentifier, resourceParameters);
                    break;
               default: throw new ProcessingException("ResourceConnector: type " +
                          resourceType + " not supported.");
            }
        } catch (SAXException local) {
            throw new ProcessingException("SAXException: " + local, local);
        } catch (IOException local) {
            throw new ProcessingException("IOException: " + local, local);
        }
        return content;
    }

    /**
     * Save a String to a resource.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     * @param content             The data to be saved.
     */
    public void saveContent(int                resourceType,
                            Parameters         typeParameters,
                            String             resourceIdentifier,
                            SourceParameters resourceParameters,
                            String             content)
    throws ProcessingException {
        try {
            switch (resourceType) {
                case ResourceConnector.RESOURCE_TYPE_URI:
                    this.saveContentToURI(typeParameters,
                                          resourceIdentifier,
                                          resourceParameters,
                                          content);
                    break;
               case ResourceConnector.RESOURCE_TYPE_CLASS:
                    throw new ProcessingException("saveContent: Type class not implemented.");
               case ResourceConnector.RESOURCE_TYPE_FILE:
                    this.saveContentToFile(typeParameters,
                                           resourceIdentifier,
                                           resourceParameters,
                                           content);
                    break;
               default: throw new ProcessingException("ResourceConnector: type " +
                          resourceType + " not supported.");
            }
        } catch (SAXException local) {
            throw new ProcessingException("SAXException: " + local, local);
        } catch (IOException local) {
            throw new ProcessingException("IOException: " + local, local);
        }
    }
}
